import json

from .path import abs_path
from .files import verify_file

class SrbJson:
    '''
    A json class with common functionality.
    parameters:
        file_path - name/path of file
        template  - a basic template for file.
                    can be of two forms:
                        1. {}   # default
                        2. { 'app_name':{} } # default embeded in something, helps to check if file is good/bad
                        3. [] for list type jsons
        strict - strict json follows template strictly and user can only modify values not keys
    '''
    def __init__(self,file_path,template={},strict=False):
        self.strict = strict
        if(type(template) is list and strict):
            raise Exception('strict mode only for dictionary rooted jsons not for list rooted jsons')
        self.file_path = abs_path(file_path)
        self.template = template
        self.masterkey = SrbJson._get_master_key(template) # generlly 'app_name'
        self.fetch_data()

    srblib_template = {
        "srblib": # masterkey
        {
            "debug":False,
        }
    }

    def fetch_data(self):
        self.data = SrbJson.extract_data(self.file_path,self.template) # also creates empty if not there
        if self.strict:
            template = self.template
            if(self.masterkey): template = template[self.masterkey]
            for key in template.keys():
                if not key in self.data:
                    raise Exception('key ' +key+ ' missing, json should be similar to template in strict mode')
            for key in self.data:
                if not key in template:
                    raise Exception('Extra key present ' +key+ ', json should be similar to template in strict mode')

        return self.data

    def keys(self):
        return self.data.keys()

    def get(self,index,notfound=None):
        try:
            return self[index]
        except:
            if notfound:
                self[index] = notfound
                return notfound
            else:
                return None

    def __getitem__(self,index):
        notfound = None
        orig_index = index
        if(type(orig_index) is tuple):
            index = orig_index[0]
            notfound = orig_index[1]

        self.fetch_data()
        if index in self.data:
            return self.data[index]

        if type(orig_index) is tuple:
            self[index] = notfound
            return notfound

        if self.masterkey and index in self.template[self.masterkey]:
            self[index] = self.template[self.masterkey][index]
        if not self.masterkey and index in self.template:
            self[index] = self.template[index]

        return self.data[index]


    def __setitem__(self,index,value):
        if self.strict:
            if index in self.data:
                self.data[index]=value
            else:
                raise Exception('cannot add new key ' +index+ ' in strict mode')
        else:
            self.data[index]=value
        self._burn_data_to_file()

    def append(self,item):
        if not type(self.data) is list:
            raise Exception(' append method only for list type jsons')
        self.data.append(item)
        self._burn_data_to_file()

    def __str__(self):
        return json.dumps(self.data, sort_keys=True, indent=4)

    def __len__(self):
        return len(self.data)

    def __contains__(self,value):
        return value in self.data

    def __delitem__(self,index):
        '''
        deletion is not allowed in strict mode. just modification is permitted
        better to set value to None
        to preserve template structure
        '''
        if self.strict:
            self[index] = None
            return

        del self.data[index]
        data = self.data
        if(self.masterkey):
            data = {self.masterkey:data}

        jfile = open(self.file_path, 'w')
        json.dump(data,jfile,indent = 4,ensure_ascii = False)
        jfile.close()


    @staticmethod
    def extract_data(file_path,template={}):
        """
        Extracts json data from the given file
        if there is no such file
            it will create one
        if there is currupt file (only for srb standard templates)
            it will create new
        if file is ok
            it will return its content
        """
        fille = abs_path(file_path)
        try:
            jfile = open(fille)
        except FileNotFoundError:
            SrbJson._create_file(fille,template)
        jfile = open(fille)
        data = json.load(jfile)

        masterkey = SrbJson._get_master_key(template) # if template is in srb standard
        if(masterkey):
            if(not masterkey in data.keys()):
                SrbJson._create_file(fille,template)
                jfile = open(fille)
                data = json.load(jfile)
            return data[masterkey]
        else:
            return data

    def _burn_data_to_file(self):
        """
        burns data to original file
        it is insecure way should be avoided to be directly used by user

        """
        data = self.data
        if(self.masterkey):
            data = {self.masterkey:data}

        jfile = open(self.file_path, 'w')
        json.dump(data,jfile,indent = 4,ensure_ascii = False)
        jfile.close()


    @staticmethod
    def dump_data(data,file_path,template):
        """
        adds data to original file
        it is bit secure, doesn't add extra key to data if not present in jfile already
        """
        temp = SrbJson(file_path,template)
        for key in data:
            if(key in temp):
                temp.data[key] = data[key]
        temp._burn_data_to_file() # lazy burning


    @staticmethod
    def _create_file(fille,template):
        verify_file(fille)
        jfile = open(fille, 'w')
        json.dump(template,jfile,indent = 4,ensure_ascii = False)
        jfile.close()

    @staticmethod
    def _get_master_key(template):
        if type(template) is list: # no master key for list
            return None
        keys = template.keys() # if template is in srb standard
        if(len(keys) == 1 and type(template[list(keys)[0]]) == dict):
            return list(keys)[0]
        else:
            return None



import os

from .srb_json import SrbJson
on_srbpc = False
if(os.environ.get('USER') == 'srb'):
    on_srbpc = True

_debug_template = {'srblib':{'debug':False,'on_travis':False,'on_appveyor':False,'on_ci':False}}
_debug_json = SrbJson('~/.config/srblib/debug.json',_debug_template)

on_travis = False
if(os.environ.get('TRAVIS','False').lower() == 'true'):
    on_travis = True
if(not on_travis):
    on_travis = _debug_json['on_travis']

on_appveyor = False
if(os.environ.get('APPVEYOR','False').lower() == 'true'):
    on_appveyor = True
if(not on_appveyor):
    on_appveyor = _debug_json['on_appveyor']

on_ci = False
if(os.environ.get('CI','False').lower() == 'true'):
    on_ci = True
if(not on_ci):
    on_ci = _debug_json['on_ci']

if on_ci:
    on_travis = True
    on_appveyor = True

debug = _debug_json['debug']




class FrozenClass(object):
    __isfrozen = False
    def __setattr__(self, key, value):
        if self.__isfrozen and not hasattr(self, key):
            raise TypeError( "%r is a frozen class" % self )
        object.__setattr__(self, key, value)

    def _freeze(self):
        self.__isfrozen = True

"""
# Turorial

class Test(FrozenClass):
    def __init__(self):
        self.x = 42#
        self.y = 2**3

        self._freeze() # no new attributes after this point.

a,b = Test(), Test()
a.x = 10
b.z = 10 # fails

"""



import os
import json
import csv

from terminaltables import AsciiTable
import xlwt # only writes into excel workbook
import xlrd # only reads from excel workbook

from .srb_json import SrbJson
from .path import abs_path
from .files import verify_file, file_extension, remove
from .util import top


class Tabular:
    def __init__(self,data=None):
        self.matrix = []
        self.json = []
        self.json_str = ""
        self.header = []
        self._parse(data)

    '''
    loader methods, each of them load matrix then call _sync()
    '''
    def load_json(self,inp_path,header=None):
        inp_path = abs_path(inp_path)
        if(not os.path.exists(inp_path)):
            raise Exception('missing input csv file')
        data = SrbJson(inp_path).data
        data = list(data)
        self.matrix = Tabular.json_to_matrix(data,header)
        self._sync()


    def load_csv(self,inp_path,delimit=','): # return tabular
        inp_path = abs_path(inp_path)
        if(not os.path.exists(inp_path)):
            raise Exception('missing input csv file')
        self.matrix = []

        inp = open(inp_path)
        inp = csv.reader(inp,delimiter=delimit)
        self.matrix = list(inp)
        self._sync()

    def load_xls(self,inp_path,strict = False):
        inp_path = abs_path(inp_path)
        if(not os.path.exists(inp_path)):
            raise Exception('missing input excel file')
        self.matrix = []

        sheet = xlrd.open_workbook(inp_path).sheets()[0]
        for i in range(sheet.nrows):
            if strict and sheet.cell_type(i,0) in (0,6):
                continue
            row = []
            sheet_row = sheet.row_values(i)
            for j in range(len(sheet_row)):
                if sheet.cell_type(i,j) in (0,6): row.append(None)
                else: row.append(sheet_row[j])
            self.matrix.append(row)
        for i in range(sheet.ncols):
            if not self.matrix[0][i]:
                raise Exception('header should not be none')
        self._sync()


    def load_matrix(self,matrix):
        refined_matrix = []
        cols = 0
        for row in matrix:
            cols = max(len(row),cols)

        for row in matrix:
            refined_row = []
            for i in range(cols):
                elem = row[i] if i < len(row) else None
                refined_row.append(elem)
            refined_matrix.append(refined_row)

        self.matrix = refined_matrix
        self._sync()


    def _sync(self):
        self.json = Tabular.matrix_to_json(self.matrix)
        self.json_str = json.dumps(self.json, sort_keys=True, indent=4)



    '''
    writer methods
    '''
    def write_json(self,out_path):
        json = SrbJson(out_path,[]) # create a SrbJson Object
        json.data = self.json
        json._burn_data_to_file() # force update

    def write_csv(self,out_path):
        out_path = abs_path(out_path)
        verify_file(out_path)
        out = open(out_path, 'w')
        out = csv.writer(out)
        for row in self.matrix:
            out.writerow(row)

    def write_xls(self,out_path):
        out_path = abs_path(out_path)
        if os.path.isfile(out_path): remove(out_path) # remove an existing one
        verify_file(out_path)
        book = xlwt.Workbook(encoding="utf-8")
        sheet = book.add_sheet("Sheet1")
        for i in range(len(self.matrix)):
            row = self.matrix[i]
            for j in range(len(row)):
                sheet.write(i,j,row[j])

        book.save(out_path)



    '''
    helper methods used for conversion
    '''
    @staticmethod
    def matrix_to_json(data): # return list of dictonaries
        header = data[0] # get header
        data = data[1:] # rest is data

        json_data = []
        for row in data:
            item = {}
            for index in range(len(header)):
                item[header[index]] = row[index]
            json_data.append(item)

        return json_data

    @staticmethod
    def json_to_matrix(data,header=None): # provide header to specify order else it will be random
        if(not header):
            first_item = data[0]
            header = list(first_item.keys()) # random order header

        if(type(header) != list):
            raise Exception('required arguments as list')

        ans = []
        ans.append(header)

        for row in data:
            row_vals = []
            for key in header:
                row_vals.append(row[key])
            ans.append(row_vals)
        return ans



    '''
    other operators and functions
    '''
    def __getitem__(self,index):
        if type(index) is str:
            loc = self.matrix[0].index(index)
            ans = []
            for i in range(1,len(self.matrix)):
                ans.append(self.matrix[i][loc])
            return ans
        return Tabular.row(self.matrix[index],self.matrix[0])

    def __len__(self):
        return len(self.matrix)

    def append(self,item):
        self.matrix.append(item)

    def __str__(self):
        return AsciiTable(self.matrix).table
        return self.matrix.__str__() # simple way



    '''
    parsers
    '''

    def _parse(self,data):
        if not data:
            return
        if type(data) is str:
            self._parse_file(data)
            return

        ans = []
        first_row = top(data)
        if type(first_row) is not dict:
            for row in data:
                ans.append(list(row))
            self.load_matrix(ans)
            return

        header = list(first_row.keys()) # random order header
        ans.append(header)
        for row in data:
            row_vals = []
            for key in header:
                row_vals.append(row[key])
            ans.append(row_vals)
        self.load_matrix(ans)

    def _parse_file(self,file_path):
        file_path = abs_path(file_path)
        ext = file_extension(file_path)
        if ext in ['csv']:
            self.load_csv(file_path)
        elif ext in ['json']:
            self.load_json(file_path)
        elif ext in ['xls','xlsx']:
            self.load_xls(file_path)
        else:
            raise Exception('Unknown file extension')



    '''
    helper class that enables named indexing into data
    '''
    class row:
        def __init__(self,data,header):
            self.header = header
            self.row = data
        def __getitem__(self,index):
            if type(index) is str:
                loc = self.header.index(index)
                return self.row[loc]
            return self.row[index]
        def __len__(self):
            return len(self.row)
        def __str__(self):
            return self.row.__str__()
            # parse items




import os
import shutil
from importlib.util import spec_from_file_location, module_from_spec

from .path import abs_path
from .files import verify_file

# ignore_var_list = ['__class__','__weakref__','__dir__']
class Module:
    def __init__(self,file_path,backup=None):
        '''
        backup should be absolute_path_of_that_file
        '''
        file_path = abs_path(file_path)
        if(not os.path.isfile(file_path)):
            if(not backup): raise('File doesnot exist please provide backup')
            backup = abs_path(backup)
            verify_file(file_path)
            shutil.copy(backup,file_path)

        mod_name = file_path.split('/')[-1].split('.')[0]
        spec = spec_from_file_location(mod_name,file_path)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)

        list_vars = dir(module)
        for var in list_vars:
            # if var not in ignore_var_list:
            setattr(self,var,getattr(module,var))




import platform
import sys

def _get_os_name():
    os_name = platform.system().lower()
    if 'window' in os_name:
        return 'windows'
    if 'darwin' in os_name:
        return 'mac'
    if 'linux' in os_name:
        return 'linux'
    return None

ok = _get_os_name() != 'windows'

class Colour:
    BOLD = '\033[1m' if ok else ''
    UNDERLINE = '\033[4m' if ok else ''
    END = '\033[0m' if ok else ''

    GRAY = '\033[90m' if ok else ''
    RED = '\033[91m' if ok else ''
    GREEN = '\033[92m' if ok else ''
    YELLOW = '\033[93m' if ok else ''
    BLUE = '\033[94m' if ok else ''
    PURPLE = '\033[95m' if ok else ''
    CYAN = '\033[96m' if ok else ''
    WHITE = '\033[97m' if ok else ''

    DARKGRAY = '\033[30m' if ok else ''
    DARKRED = '\033[31m' if ok else ''
    DARKGREEN = '\033[32m' if ok else ''
    DARKYELLOW = '\033[33m' if ok else ''
    DARKBLUE = '\033[34m' if ok else ''
    DARKPURPLE = '\033[35m' if ok else ''
    DARKCYAN = '\033[36m' if ok else ''
    DARKWHITE = '\033[37m' if ok else ''

    FULLDARKGRAY = '\033[40m' if ok else ''
    FULLDARKRED = '\033[41m' if ok else ''
    FULLDARKGREEN = '\033[42m' if ok else ''
    FULLDARKYELLOW = '\033[43m' if ok else ''
    FULLDARKBLUE = '\033[44m' if ok else ''
    FULLDARKPURPLE = '\033[45m' if ok else ''
    FULLDARKCYAN = '\033[46m' if ok else ''
    FULLDARKWHITE = '\033[47m' if ok else ''

    FULLGRAY = '\033[100m' if ok else ''
    FULLRED = '\033[101m' if ok else ''
    FULLGREEN = '\033[102m' if ok else ''
    FULLYELLOW = '\033[103m' if ok else ''
    FULLBLUE = '\033[104m' if ok else ''
    FULLPURPLE = '\033[105m' if ok else ''
    FULLCYAN = '\033[106m' if ok else ''
    FULLWHITE = '\033[107m' if ok else ''


    def print(message,colour='',end='\n'):
        os_name = _get_os_name()
        end_colour = Colour.END
        if(os_name == 'windows'):
            colour = ''
            end_colour = ''
        # print(colour + message + end_colour,end=end)
        sys.stderr.write(colour + message + end_colour + end)




#! /usr/bin/env python3
import hashlib
import os
import sys
import time

from .path import abs_path

def _hash_file(fname):
    fname = abs_path(fname)
    BLOCKSIZE = 65536
    hasher = hashlib.md5()
    with open(fname, 'rb') as afile:
        buf = afile.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(BLOCKSIZE)
    return hasher.hexdigest()


def _dfs_dir(path):
    content = os.listdir(path)
    dir_hash = "+"
    for a in content:
        a = os.path.join(path,a)
        if os.path.isdir(a):
            val = _dfs_dir(a)
            dir_hash += val
        elif os.path.isfile(a):
            dir_hash += '-'
            dir_hash += str(_hash_file(a))
    return dir_hash


def str_hash(inp):
    inp = str(inp)
    return str(hashlib.md5(inp.encode('utf-8')).hexdigest())

def path_hash(path):
    path = abs_path(path)
    if not os.path.exists(path):
        raise Exception('Path not found ' + path)
    if os.path.isfile(path):
        return _hash_file(path)
    dir_hash = _dfs_dir(path)
    return str_hash(dir_hash)



import platform

def get_os_name():
    os_name = platform.system().lower()
    if 'window' in os_name:
        return 'windows'
    if 'darwin' in os_name:
        return 'mac'
    if 'linux' in os_name:
        return 'linux'
    return None

os_name = get_os_name()

on_windows = False
if os_name == 'windows':
    on_windows = True



import os
import subprocess

from .util import line_adder, dump_output
from .colour import Colour

def _get_supported_distros(dependency_map):
    supported_distros = set()
    for rules in dependency_map.values():
        for key in rules.keys():
            supported_distros.add(key)
    return supported_distros


def _recognise_distro(distros=['ubuntu']):
    try:
        p = subprocess.Popen(['uname','-a'], stdout=subprocess.PIPE)
        out = p.stdout.read().decode('utf-8').lower()
        for d in distros:
            if d in out:
                return d
        return None
    except:
        return None

def _get_installed_package_mangers(available_package_managers):
    installed_package_managers = set()
    for apm in available_package_managers:
        if is_installed(apm):
            installed_package_managers.add(apm)
    return installed_package_managers


def is_installed(soft):
    help_opt = ' --help '
    a = os.system(soft + help_opt + dump_output)
    if a == 0 or a == 256:
        '''
        0 means return 0
        256 means return 1, generally for those who dont have --help
        if command not found it will return 32512
        '''
        return True
    return False


def install_arg_complete(app_name):
    if is_installed('register-python-argcomplete'):
        line = 'eval "$(register-python-argcomplete '+app_name+')"'
        filename = os.environ['HOME'] + '/.bashrc'
        line_adder(filename,line)


def install_dependencies(dependency_map, verbose = False):
    supported_distros = _get_supported_distros(dependency_map)
    distro = _recognise_distro(supported_distros)
    if(verbose):
        if(not distro):
            Colour.print('unrecognised distro, please contact srbcheema2@gmail.com for support', Colour.RED)
        else:
            Colour.print('Distro detected to be '+distro+' based',Colour.GREEN)

    all_installed = True

    for d in dependency_map.keys():
        if is_installed(d):
            continue
        rules = dependency_map[d]
        if distro and distro in rules.keys():
            Colour.print('installing '+d+' dependency',Colour.GREEN)
            os.system(rules[distro])
            if not is_installed(d):
                Colour.print('please install ' +d+ ' dependency manually',Colour.YELLOW)
                Colour.print('try command : '+rules[distro],Colour.YELLOW)
                all_installed = False
        else:
            Colour.print('Please install ' +d+ ' dependency manually',Colour.YELLOW)
            all_installed = False

    return all_installed

def remove_dependencies(dependency_map, verbose = False):
    supported_distros = _get_supported_distros(dependency_map)
    distro = _recognise_distro(supported_distros)
    if(verbose):
        if(not distro):
            Colour.print('unrecognised distro, please contact srbcheema2@gmail.com for support', Colour.RED)
        else:
            Colour.print('Distro detected to be '+distro+' based',Colour.GREEN)

    all_removed = True

    for d in dependency_map.keys():
        if not is_installed(d):
            continue
        rules = dependency_map[d]
        if distro and distro in rules.keys():
            Colour.print('removing '+d+' dependency',Colour.GREEN)
            os.system(rules[distro])
            if is_installed(d):
                Colour.print('please remove ' +d+ ' dependency manually',Colour.YELLOW)
                Colour.print('try command : '+rules[distro],Colour.YELLOW)
                all_removed = False
        else:
            Colour.print('Please remove ' +d+' dependency manually',Colour.YELLOW)
            all_removed = False

    return all_removed

def install_dependencies_pkg(dependency_map, verbose=False):
    all_installed = True
    for dependency in dependency_map.keys():
        if is_installed(dependency):
            if verbose: Colour.print('.:Dependency '+dependency+' already installed...',Colour.GREEN)
            continue

        rules = dependency_map[dependency]
        installed_package_managers = _get_installed_package_mangers(rules.keys())

        if len(installed_package_managers) == 0:
            Colour.print('No supported package manager available for ' +Colour.END+ dependency +Colour.RED
                    + ' please contact srbcheema2@gmail.com for support', Colour.RED)
            continue

        Colour.print('.:Installing '+dependency+' dependency',Colour.GREEN)
        for ipm in installed_package_managers :
            os.system(rules[ipm])
            if is_installed(dependency) :# else try other package managers
                break
        if not is_installed(dependency):
            Colour.print('please install ' +Colour.END+ dependency +Colour.YELLOW+ ' manually',Colour.YELLOW)
            for ipm in installed_package_managers :
                Colour.print('try command : '+Colour.END+ rules[ipm],Colour.YELLOW)
            all_installed = False
    return all_installed

def remove_dependencies_pkg(dependency_map, verbose=False):
    all_removed = True
    for dependency in dependency_map.keys():
        if not is_installed(dependency):
            if verbose: Colour.print('.:Dependency '+dependency+' already not installed...',Colour.GREEN)
            continue

        rules = dependency_map[dependency]
        installed_package_managers = _get_installed_package_mangers(rules.keys())

        if len(installed_package_managers) == 0:
            Colour.print('No supported package manager rule available for ' +Colour.END+ dependency + Colour.RED
                    + ' please contact srbcheema2@gmail.com for support', Colour.RED)
            continue

        Colour.print('.:Uninstalling '+dependency+' dependency',Colour.GREEN)
        for ipm in installed_package_managers :
            os.system(rules[ipm])
            if is_installed(dependency) :# else try other package managers
                break
        if not is_installed(dependency):
            Colour.print('please uninstall ' +Colour.END+ dependency +Colour.YELLOW+ ' manually',Colour.YELLOW)
            for ipm in installed_package_managers :
                Colour.print('try command : '+Colour.END+ rules[ipm],Colour.YELLOW)
            all_removed = False
    return all_removed



if __name__ == '__main__':
    dependency_map = {
        'register-python-argcomplete':{
            'ubuntu':'sudo apt remove python-argcomplete',
        },
        'figlet':{
            'ubuntu':'sudo apt remove figlet',
        },
    }
    remove_dependencies(dependency_map,verbose=True)
    dependency_map = {
        'register-python-argcomplete':{
            'ubuntu':'sudo apt install python-argcomplete',
        },
        'figlet':{
            'ubuntu':'sudo apt install figlet',
        },
    }
    install_dependencies(dependency_map,verbose=True)
    install_arg_complete('demo_app')



from bs4 import BeautifulSoup
import requests

from .colour import Colour

class Soup:
    def __init__(self,inp):
        if type(inp) is str:
            self.soup = Soup.get_soup(inp)
        elif type(inp) is BeautifulSoup:
            self.soup = inp
        elif type(inp) is Soup._trusted:
            self.soup = inp._data
        else:
            raise Exception('requires string as argument')

        self._sync()

    def __getitem__(self,index):
        if(type(index) is tuple):
            return self.find_all(*index)
        return self.find_all(index)

    def find_all(self,*args,**kwargs):
        ret = self.soup.find_all(*args,**kwargs)
        return Soup._SO(ret)

    def __getattr__(self, name): # this one is better way
        try:
            return getattr(self,name)
        except KeyboardInterrupt:
            raise KeyboardInterrupt
        except:
            return getattr(self.soup,name)

    def __str__(self):
        return self.soup.prettify()

    def _sync(self):
        if self.soup.parent is None:
            self.parent = None
        else:
            self.parent = Soup(Soup._trusted(self.soup.parent))

    @staticmethod
    def get_soup(url,*args,**kwargs):
        '''
        takes url and return soup.
        returns None if there is bad connection or bad response code.
        '''
        result = requests.get(url, *args, **kwargs)
        if(result is None or result.status_code is not 200):
            if(result == None):
                temp_result = requests.get('https://google.com')
                if(temp_result == None):
                    colour.print('please check your internet connection', colour.red)
                else:
                    colour.print('please check your url', colour.red)
            else:
                Colour.print('soup result on '+url+' :'+Colour.END+str(result), Colour.RED)
            return None
        soup = BeautifulSoup(result.text, 'html.parser')
        return soup

    class _SO:
        def __init__(self,inp):
            if type(inp) is Soup._trusted:
                self._data = inp._data
                return
            self._data = inp # it is actually a list

        def __getitem__(self,index):
            if(type(index) is slice): # a[1:9]
                return Soup._SO(Soup._trusted(self._data[index]))
            return Soup(Soup._trusted(self._data[index]))

        def __len__(self):
            return len(self._data)

        def __str__(self):
            return self._data.__str__()

    class _trusted:
        def __init__(self,_data):
            self._data = _data



import os
import shutil

from .path import abs_path
from .colour import Colour


def verified_file(file_path):
    '''
    returns abs_path if exists or raise exception
    '''
    file_path = abs_path(file_path)
    if not os.path.isfile(file_path):
        raise Exception('No such file : ' + file_path)
    return file_path

def verify_folder(folder,debug=False):
    '''
    similar to mkdir -p
    '''
    folder = abs_path(folder)
    if not os.path.exists(folder):
        if(debug): print('creating folder '+ folder)
        os.makedirs(folder)
    elif os.path.isfile(folder):
        if(debug): print('there exists file of same name')

def verify_file(file_path,debug=False):
    '''
    verifies a file. if it doesn't exists creates one
    '''
    file_path = abs_path(file_path)
    parent_dir = os.path.join(file_path,os.pardir)
    verify_folder(parent_dir)
    if not os.path.exists(file_path):
        if(debug): print('creating file '+ file_path)
        file_ = open(file_path, 'w')
        file_.close()
    elif os.path.isdir(file_path):
        if(debug): print('there exists folder of same name')

def remove(path,debug=False):
    '''
    removes a folder/file of given path.
    works recursively for folders
    '''
    path = abs_path(path)
    if not os.path.exists(path):
        return
    if os.path.isfile(path):
        os.remove(path)
    elif os.path.isdir(path):
        shutil.rmtree(path)
    else:
        Colour.print('unknown path : '+path,Colour.RED)



def file_name(file_path):
    file_path = abs_path(file_path)
    items = file_path.split('/')[-1]
    return items

def file_extension(file_path):
    file_name_ = file_name(file_path)
    if(file_name_[0] == '.'): file_name_ = file_name_[1:]
    if('.' in file_name_):
        return file_name_.split('.')[-1]
    else:
        return ''



import sys

try:
    import argparse
    from argcomplete import autocomplete
except:
    from .util import show_dependency_error_and_exit
    show_dependency_error_and_exit()

from . import __mod_name__,__version__
from .debugger import _debug_json
from .colour import Colour

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version",
                        action='store_true',
                        help="Display version number")
    parser.add_argument("-d", "--debug",
                        action='store_true',
                        help="Toggle value of debug variable")
    autocomplete(parser)
    parsed_args = parser.parse_args()
    if(parsed_args.version):
        print(__mod_name__+'=='+__version__)
        sys.exit()
    if(parsed_args.debug):
        _debug_json['debug'] = not _debug_json['debug']
        Colour.print('debug value set to: ' + str(_debug_json['debug']), Colour.GREEN)
    print(__file__)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import os
from setuptools import setup, find_packages

from srblib import __version__, __mod_name__

with open("README.md", 'r') as f:
    long_description = f.read()
with open('requirements.txt', 'r') as f:
    requirements = [line.strip() for line in f.readlines()]

setup(
    name=__mod_name__,
    version=__version__,
    description='A python library by srb',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Sarbjit Singh',
    author_email='srbcheema1@gmail.com',
    url='http://github.com/srbcheema1/'+__mod_name__,

    packages=find_packages(), # provides same list, looks for __init__.py file in dir
    include_package_data=True,
    install_requires=requirements, #external packages as dependencies

    entry_points={
        'console_scripts': [__mod_name__+'='+__mod_name__+'.main:main']
    },

    classifiers=[
        'Operating System :: POSIX :: Linux',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
    ],
    license='MIT License',
)



__version__ = '0.1.0'
__mod_name__ = 'srblib'


from .colour import Colour # A class with color names and a static print function which prints coloured output to stderr

from .debugger import debug # a boolean whose value can be changed in ~/.config/srblib/debug.json
from .debugger import on_appveyor # a boolean value which is true if code is running on appveyor
from .debugger import on_ci # a boolean value which is true if it code is running on CI
from .debugger import on_srbpc # a boolean value which is true if it is my PC i.e. srb-pc
from .debugger import on_travis # a boolean value which is true if code is running on travis

from .dependency import install_arg_complete # A function to append line of argcomplete in ~/.bashrc
from .dependency import install_dependencies # A function that takes a special data-template to install dependencies
from .dependency import install_dependencies_pkg # similar but based on package-managers (Recommended)
from .dependency import is_installed # checks if the following application is installed on machine or not
from .dependency import remove_dependencies # Opposite of install_dependencies
from .dependency import remove_dependencies_pkg # Opposite of install_dependencies_pkg

from .files import file_extension # returns back the extention of a file from filepath, may return '' if no ext
from .files import file_name # returns filename from a filepath
from .files import remove # removes a path recursively. it deletes all files and folders under that path
from .files import verify_file # verify that a file exists. if not it will create one. also creates parents if needed
from .files import verify_folder # verify that a folder exists. creates one if not there. also creates parents if needed

from .file_importer import Module # a class to import modules

# one cant declare more attributes in frozen class
from .frozen import FrozenClass # A class to be inherited to make a class frozen. i.e. no more attributes can be added.

from .path import abs_path # returns absolute path of a path given. works on windows as well as linux.
from .path import is_child_of # returns if a given path is child(direct/indirect) of the second path given.
from .path import parent_dir # returns Nth parent of a path. default it returns 1st parent
from .path import relative_path # returns relative path if given absolute path

from .srb_bank import SrbBank # A class to store things for later use of a program. can act as a database
from .srb_json import SrbJson # A class to use json file more easily
from .srb_hash import path_hash # get hash of full path (recursively)
from .srb_hash import str_hash # get hash of string

from .soup import Soup # A class to make scrapping easier

from .system import get_os_name # returns OS name. values are windows, linux or mac
from .system import os_name # value of get_os_name
from .system import on_windows # True if system is windows OS

from .tabular import Tabular # A class to process dabular data

from .util import line_adder # append a line if not present in a given file
from .util import show_dependency_error_and_exit # display missing dependency error and exit
from .util import similarity # returns percentage of similarity of two strings
from .util import top # first element of list or set or dict(first key)
from .util import dump_output # variable containing string value ` > /dev/null 2>&1 ` or ` > nul 2>&1 `.



#!/usr/bin/env python3
import os
from sys import argv

from .system import get_os_name

def _reactified(path):
    '''
    assumes that path has forward slashes in it
    and no slash at the end
    '''
    drive = None
    nodes = path.split("/")
    if(nodes[0] != ''): drive = nodes[0] # for windows it will pick C: or D:

    direc = []
    nodes = nodes[1:] # remove everthing before first slash, its contained in drive
    for node in nodes:
        if(node=='.'): continue
        elif(node=='..'):
            if(len(direc)>0):
                direc.pop()
            else:
                raise Exception("Cant go beyond base directory, arg: "+path)
        else: direc.append(node)

    path = ""
    for item in direc:
        path = path + "/"
        path = path + item

    if(drive): path = drive + path # add drive again
    return path


def _abs_path_unix(path):
    pwd = str(os.getcwd())#always return without / at end;
    home_path = os.getenv("HOME")#always return without / at end
    if(not home_path): raise Exception('home path variable not set')

    if(len(path)>1 and path[-1]=='/'):#remove last backslash
        path=path[:-1]
    if(path[0]=='/'):
        return _reactified(path)
    if(path[0]=='~'):
        return _reactified(home_path + path[1:])
    if(path[0]!='.'):
        return _reactified(pwd + '/' + path)

    return _reactified(pwd +'/'+ path)



def _abs_path_windows(path,debug=False):
    '''
    takes path in any slash but return only in forward slash
    '''
    pwd = str(os.getcwd())#always return without / at end except drives ex C:\
    home_path = os.getenv("userprofile")#always return without / at end, ex: C:\User\srb
    if(not home_path):
        if(debug): home_path = 'C:\\User\\testing'
        else: raise Exception('home path variable not set')

    # make paths better looking
    pwd = _change_slash(pwd)
    pwd = _reactified(pwd)
    home_path = _change_slash(home_path)
    path = _change_slash(path)
    if(len(path)>1 and path[-1]=='/'):#remove last backslash
        path=path[:-1]

    if(path[0]=='/'):
        raise Exception('windows path cant start with slash')
    if(path[0]=='~'):
        ans = _reactified(home_path + path[1:])
    elif(len(path)>1 and path[1]==':'): #path is of type C:.... or D:.....
        ans = _reactified(path)
    else:
        ans = _reactified(pwd +'/'+ path)

    return ans

def _change_slash(path,slash='/'):
    ans = ""
    for a in path:
        if a in ['/','\\']:
            ans+=slash
        else:
            ans+=a
    return ans

'''
takes relative path as string argument as a path in linux
returns absolute path without '/' at end
means it will write /home/srb instead of /home/srb/

its more helpful when dealing with paths of files.
both the paths will be in sync as both wont have / at end
'''

def abs_path(path,slash='/'):
    os_name = get_os_name()
    if(os_name=='windows'):
        path = _abs_path_windows(path)
        return _change_slash(path,slash)
    else:
        return _abs_path_unix(path)

def is_child_of(parent,child):
    '''
    detects if child path is child of parent path
    '''
    parent = abs_path(parent)
    child = abs_path(child)
    if(parent in child): return True
    return False


def parent_dir(file_path,level=1):
    here = str(abs_path(file_path))
    while(level>0):
        here = abs_path(os.path.join(here, os.pardir))
        level -= 1
        if(here==''):
            return '/'

    return abs_path(here)

def relative_path(file_path):
    file_path = abs_path(file_path)
    pwd = abs_path(os.getcwd())
    file_path = file_path[len(pwd):len(file_path)]
    return '.'+file_path

if __name__ == "__main__":
    '''
    point to be noted:
    if you run as ./abs_path.py hello\world
    you will get different values in argv[1]
    windows:
        hello\world
    unix:
        helloworld
    '''
    print(argv[1])
    print(abs_path(argv[1],'\\'))



import getpass
import hashlib
import os
import pickle
import json

from .path import abs_path
from .files import verify_file, file_name, file_extension

class SrbBank:
    '''
    args:
        filename:- filename to be used to store data, it will be a file locally without .bank extension
        password:- password if needed. to open a passwrd protected file you need to send password
    '''
    def __init__(self,filename="temp",password=None):
        self._password = password
        self._loggedin = False
        self._data = dict()

        filename = abs_path('.'+filename)
        file_ext = file_extension(filename)
        if file_ext == '':
            filename += '.bank'
        else:
            raise Exception('file_name should not have extension')
        self._file_name = filename
        if not os.path.exists(self._file_name):
            self._loggedin = True
            self._save() # needs loggedin

        self._login(self._password)
        self.load()


    def _login(self,password):
        '''
        logins into a bank account
        once logged in we can perform read write operations
        '''
        data = pickle.load(open(self._file_name,'rb'))
        _password = data.get('password')
        if(not _password): # not password protected
            self._password = None
            self._loggedin = True
            return

        if(not password): raise Exception('password protected, please provide password')

        if(SrbBank._md5(password) == _password): self._loggedin = True
        else: raise Exception('wrong password')


    def _authenticator(func):
        def wrapper(self, *args, **kargs):
            if not self._loggedin:
                raise Exception('Function only available when logged in')
            return func(self, *args, **kargs)
        return wrapper

    @_authenticator
    def __str__(self):
        return json.dumps(self._data, sort_keys=True, indent=4)

    @_authenticator
    def setpass(self,password=None):
        if(password):
            self._password = SrbBank._md5(password)
        else:
            while True:
                password = getpass.getpass('Enter your password:')
                _pasword = getpass.getpass('ReEnter your password:')
                if(password == _pasword):
                    self._password = SrbBank._md5(password)
                    break
                else:
                    print('passwords do not match')
        self._save()

    @_authenticator
    def removepass(self):
        self._password = None
        self._save()


    @_authenticator
    def load(self):
        data = pickle.load(open(self._file_name,'rb'))
        self._password = data.get('password')
        if(self._password):
            del data['password']
        self._data = data
        return data

    @_authenticator
    def __getitem__(self,index):
        if(index in self._data):
            return self._data[index]
        else:
            return None

    @_authenticator
    def __setitem__(self,index,value):
        self._data[index]=value
        self._save()

    @_authenticator
    def __delitem__(self,index):
        del self._data[index]
        self._save()

    @_authenticator
    def __contains__(self,value):
        return value in self._data

    @_authenticator
    def _save(self):
        data = self._data
        if(self._password):
            data['password'] = self._password
        pickle.dump(data,open(self._file_name,'wb'))

    @staticmethod
    def _md5(inp):
        return hashlib.md5(inp.encode('utf-8')).hexdigest()



import sys


from .colour import Colour
from .debugger import debug, on_travis
from .system import os_name


dump_output = ' > /dev/null 2>&1 '
if os_name == 'windows':
    dump_output = ' > nul 2>&1 '

_dependency_err_1 = """
    You haven't installed the required dependencies.
    Please install required dependencies using :"""
_dependency_err_2 = """        python3 -m pip install -r requirements.txt
"""
def show_dependency_error_and_exit():
    Colour.print(_dependency_err_1,Colour.RED)
    Colour.print(_dependency_err_2,Colour.YELLOW)
    if(debug or on_travis):
        import traceback
        traceback.print_exc()
    sys.exit(1)


def line_adder(filename, line):
    with open(filename, 'r+') as f:
        content = f.read()
        f.seek(0, 0)
        lines = [x.strip() for x in f.readlines()]
        if(not line in lines):
            f.seek(0, 0)
            f.write(content + '\n' + line + '\n')

def top(container):
    for item in container:
        return item
    return None


from difflib import SequenceMatcher
def similarity(a, b):
    def _lcs(a,b):
        def _lcs_rec(i=0,j=0):
            if(i==len(a) or j==len(b)): return 0
            if dp[i][j] != -1: return dp[i][j]
            if(a[i] == b[j]): return 1 + _lcs(i+1,j+1)
            ret1 = _lcs(i,j+1)
            ret2 = _lcs(i+1,j)
            dp[i][j] = max(ret1,ret2)
            return dp[i][j]
        dp = []

        for i in range(len(a)):
            arr = []
            for j in range(len(b)):
                arr.append(-1)
            dp.append(arr)
        return _lcs_rec(0,0)
    a = str(a).lower()
    b = str(b).lower()
    if(len(a) * len(b) >= 1000):
        if(len(a) * len(b) >= 100000): raise Exception('Too long strings')
        ret = SequenceMatcher(None, a, b).ratio() * 100
        return 100 - (100-ret)*90/100 # reduce mismatch by 10%
    lcs = _lcs(a,b)
    lcs = int(lcs / min(len(a),len(b)) * 100)
    ret = SequenceMatcher(None, a, b).ratio() * 100
    ret = (ret + (1.5)*lcs)/(2.5)
    ret = (ret * lcs)/100
    return int(ret)
